"""
Geometric Brownian motion module
"""
from jaxfin.models.gbm.gbm import (
    MultiGeometricBrownianMotion,
    UnivGeometricBrownianMotion,
)

__all__ = ["UnivGeometricBrownianMotion", "MultiGeometricBrownianMotion"]



"""
Utils funciton of the model module
"""
import jax
import jax.numpy as jnp


def check_symmetric(a: jax.Array, tol=1e-8):
    """
    Check if a matrix is symmetric

    :param a: (jax.Array): Matrix to check
    :param tol: (float): Tolerance for the check
    :return: (bool): True if the matrix is symmetric
    """
    return jnp.all(jnp.abs(a - a.T) < tol)



"""Geometric Brownian Motion class implementation"""
from typing import List, Optional, Union

import jax
import jax.numpy as jnp
from jax import random

from ..utils import check_symmetric


class UnivGeometricBrownianMotion:
    """
    Geometric Brownian Motion

    Represent a 1-dimensional GBM

    # Example usage:
    s0 = jnp.array([100], dtype=jnp.float32)
    mean = jnp.array([0.1], dtype=jnp.float32)
    sigma = jnp.array([0.3], dtype=jnp.float32)

    gbm_process = UnivGeometricBrownianMotion(s0, mean, sigma, dtype=jnp.float32)
    paths = gbm_process.simulate_paths(maturity=1.0, n=100, n_sim=100)
    """

    def __init__(
        self,
        s0: Union[jax.Array, float],
        mean: Union[jax.Array, float],
        sigma: Union[jax.Array, float],
        dtype: Optional[jax.numpy.dtype] = jnp.float32,
    ):
        """GBM constructor

        Args:
            s0 (Union[jax.Array, float]): The initial value of the GBM
            mean (Union[jax.Array, float]): The drift term (mean) of the GBM
            sigma (Union[jax.Array, float]): The standard deviation of the GBM
            dtype (Optional[jax.numpy.dtype], optional): The dtype of the GBM. Defaults to jnp.float32.

        Raises:
            ValueError: When the dtype is None
        """
        if dtype is None:
            raise ValueError("dtype must not be None")

        self._dtype = dtype
        self._s0 = jnp.asarray(s0, dtype=dtype)
        self._mean = jnp.asarray(mean, dtype=dtype)
        self._sigma = jnp.asarray(sigma, dtype=dtype)

    @property
    def mean(self) -> jax.Array:
        """Return the mean of the GBM

        Returns:
            jax.Array: The mean of the GBM
        """
        return self._mean

    @property
    def sigma(self) -> jax.Array:
        """Return the standard deviation of the GBM

        Returns:
            jax.Array: The standard deviation of the GBM
        """
        return self._sigma

    @property
    def s0(self) -> jax.Array:
        """Return the initial value of the GBM

        Returns:
            jax.Array: The initial value of the GBM
        """
        return self._s0

    @property
    def dtype(self) -> jax.numpy.dtype:
        """Return the underlying dtype of the GBM

        Returns:
            jax.numpy.dtype: The dtype of the GBM
        """
        return self._dtype

    def sample_paths(self, seed: int, maturity: float, n: int, n_sim: int) -> jax.Array:
        """
        Simulate a sample of paths from the Geometric Brownian Motion (GBM).

        Args:
            maturity (float): Time in years.
            n (int): Number of steps.
            n_sim (int): Number of simulations.

        Returns:
            jax.Array: Array containing the sample paths.
        """
        key = random.PRNGKey(seed)

        dt = maturity / n

        Xt = jnp.exp(
            (self._mean - self._sigma**2 / 2) * dt
            + self._sigma * jnp.sqrt(dt) * random.normal(key, shape=(n_sim, n - 1)).T
        )

        Xt = jnp.vstack([jnp.ones(n_sim), Xt])

        return self._s0 * Xt.cumprod(axis=0)


class MultiGeometricBrownianMotion:
    """
    Geometric Brownian Motion

    Represent a d-dimensional GBM

    # Example usage:
    s0 = jnp.array([100, 100], dtype=jnp.float32)
    mean = jnp.array([0.1, 0.1], dtype=jnp.float32)
    sigma = jnp.array([0.3, 0.3], dtype=jnp.float32)
    corr = jnp.array([[1.0, 0.5], [0.5, 1.0]], dtype=jnp.float32)

    m_gbm = MultiGeometricBrownianMotion(s0, mean, sigma, corr, jnp.float32)
    paths = gmb_process.simulate_paths(maturity=1.0, n=100, n_sim=100)
    """

    def __init__(
        self,
        s0: Union[jax.Array, List[float]],
        mean: Union[jax.Array, List[float]],
        sigma: Union[jax.Array, List[float]],
        corr: jax.Array,
        dtype: Optional[jax.numpy.dtype] = jnp.float32,
    ):
        """Multivariate GBM constructor

        Args:
            s0 (Union[jax.Array, List[float]]): the initial values of the GBM
            mean (Union[jax.Array, List[float]]): the mean of the GBM
            sigma (Union[jax.Array, List[float]]): the standard deviation of the GBM
            corr (jax.Array): the correlation matrix of the GBM
            dtype (Optional[jax.numpy.dtype], optional): the type of the arrays. Defaults to jnp.float32.

        Raises:
            ValueError: when the dtype is None
            ValueError: When the correlation matrix is not symmetric
            ValueError: When the correlation matrix does not have ones as diagonal elements
        """
        if dtype is None:
            raise ValueError("dtype must not be None")

        if not check_symmetric(corr, 1e-8):
            raise ValueError("Correlation matrix must be symmetric")

        if not jnp.array_equal(jnp.diag(corr), jnp.ones(corr.shape[0])):
            raise ValueError("Correlation matrix must have ones as diagonal elements")

        self._dtype = dtype
        self._s0 = jnp.asarray(s0, dtype=dtype)
        self._mean = jnp.asarray(mean, dtype=dtype)
        self._sigma = jnp.asarray(sigma, dtype=dtype)
        self._corr = jnp.asarray(corr, dtype=dtype)
        self._dim = self._s0.shape[0]

    @property
    def mean(self) -> jax.Array:
        """
        :return: Returns the mean of the GBM
        """
        return self._mean

    @property
    def sigma(self) -> jax.Array:
        """
        :return: Returns the standard deviation of the GBM
        """
        return self._sigma

    @property
    def corr(self) -> jax.Array:
        """
        :return: Returns the correlation matrix of the Weiner processes
        """
        return self._corr

    @property
    def s0(self) -> jax.Array:
        """
        :return: Returns the initial value of the GBM
        """
        return self._s0

    @property
    def dtype(self) -> jax.numpy.dtype:
        """
        :return: Returns the underlying dtype of the GBM
        """
        return self._dtype

    @property
    def dimension(self) -> int:
        """
        :return: Returns the dimension of the GBM
        """
        return self._dim

    def sample_paths(self, seed: int, maturity: float, n: int, n_sim: int) -> jax.Array:
        """
        Simulate a sample of paths from the Multivariate Geometric Brownian Motion (GBM).

        Args:
            maturity (float): Time in years.
            n (int): Number of steps.
            n_sim (int): Number of simulations.

        Returns:
            jax.Array: Array containing the sample paths.
        """
        key = random.PRNGKey(seed)

        dt = maturity / n

        normal_draw = random.normal(key, shape=(n_sim, n * self._dim))
        normal_draw = jnp.reshape(normal_draw, (n_sim, n, self._dim)).transpose(
            (1, 0, 2)
        )

        cholesky_matrix = jnp.linalg.cholesky(self._corr)

        stochastic_increments = normal_draw @ cholesky_matrix

        log_increments = (self._mean - self._sigma**2 / 2) * dt + jnp.sqrt(
            dt
        ) * self._sigma * stochastic_increments

        once = jnp.ones([n, n], dtype=self._dtype)
        lower_triangular = jnp.tril(once, k=-1)
        cumsum = log_increments.transpose() @ lower_triangular
        cumsum = cumsum.transpose((1, 2, 0))
        samples = self._s0 * jnp.exp(cumsum)
        return samples.transpose(1, 0, 2)[::-1, :, :]



"""
Heston model class implementation
"""
from typing import Optional, Union

import jax
import jax.numpy as jnp
from jax import jit, random

from ..utils import check_symmetric


class UnivHestonModel:
    """
    Univariate Heston Model

    Represent a 1-dimensional Heston model

    # Example usage:
    s0 = jnp.array([100], dtype=jnp.float32)
    v0 = jnp.array([0.04], dtype=jnp.float32)
    kappa = jnp.array([2.0], dtype=jnp.float32)
    theta = jnp.array([0.04], dtype=jnp.float32)
    sigma = jnp.array([0.3], dtype=jnp.float32)
    rho = jnp.array([-0.7], dtype=jnp.float32)

    heston_model = UnivHestonModel(s0, v0, kappa, theta, sigma, rho, dtype=jnp.float32)
    paths = heston_model.simulate_paths(maturity=1.0, n=100, n_sim=100)
    """

    def __init__(
        self,
        s0: Union[jax.Array, float],
        v0: Union[jax.Array, float],
        mean: Union[jax.Array, float],
        kappa: Union[jax.Array, float],
        theta: Union[jax.Array, float],
        sigma: Union[jax.Array, float],
        rho: Union[jax.Array, float],
        dtype: Optional[jax.numpy.dtype] = jnp.float32,
    ):
        """Univariate Heston model constructor

        Args:
            s0 (Union[jax.Array, float]): The initial value of the asset
            v0 (Union[jax.Array, float]): The initial of the variance of the asset
            mean (Union[jax.Array, float]): The mean of the asset
            kappa (Union[jax.Array, float]): The speed of the mean-reversion of the variance
            theta (Union[jax.Array, float]): The long-term mean of the variance
            sigma (Union[jax.Array, float]): The volatility of the variance
            rho (Union[jax.Array, float]): The correlation between the asset and the variance
            dtype (Optional[jax.numpy.dtype], optional): The type(jnp.float32, ...). Defaults to jnp.float32.
        """
        if dtype is None:
            raise ValueError("dtype must not be None")

        self._dtype = dtype
        self._s0 = jnp.asarray(s0, dtype=dtype)
        self._v0 = jnp.asarray(v0, dtype=dtype)
        self._mean = jnp.asarray(mean, dtype=dtype)
        self._kappa = jnp.asarray(kappa, dtype=dtype)
        self._theta = jnp.asarray(theta, dtype=dtype)
        self._sigma = jnp.asarray(sigma, dtype=dtype)
        self._rho = jnp.asarray(rho, dtype=dtype)

    @property
    def mean(self) -> jax.Array:
        """Returns the mean of the Heston model

        Returns:
            jax.Array: The mean of the Heston model
        """
        return self._mean

    @property
    def kappa(self) -> jax.Array:
        """Returns the speed of the mean-reversion of the variance

        Returns:
            jax.Array: The speed of the mean-reversion of the variance
        """
        return self._kappa

    @property
    def theta(self) -> jax.Array:
        """Returns the long-term mean of the variance

        Returns:
            jax.Array: The long-term mean of the variance
        """
        return self._theta

    @property
    def sigma(self) -> jax.Array:
        """Returns the volatility of the variance

        Returns:
            jax.Array: The volatility of the variance
        """
        return self._sigma

    @property
    def rho(self) -> jax.Array:
        """Returns the correlation between the asset and the variance

        Returns:
            jax.Array: The correlation between the asset and the variance
        """
        return self._rho

    @property
    def s0(self) -> jax.Array:
        """Returns the initial value of the asset

        Returns:
            jax.Array: The initial value of the asset
        """
        return self._s0

    @property
    def v0(self) -> jax.Array:
        """Returns the initial value of the variance of the asset

        Returns:
            jax.Array: The initial value of the variance of the asset
        """
        return self._v0

    @property
    def dtype(self) -> jax.numpy.dtype:
        """Returns the underlying dtype of the Heston model

        Returns:
            jax.numpy.dtype: The underlying dtype of the Heston model
        """
        return self._dtype

    def sample_paths(self, seed: int, maturity: float, n: int, n_sim: int):
        """
        Sample of paths from the Univariate Heston model

        Args:
            seed (int): the seed for the random number generator
            maturity (float): The time in years
            n (int): Number of steps (discretization points)
            n_sim (int): The number of simulations
        """
        key = random.PRNGKey(seed)

        dt = maturity / n

        W = random.normal(key, shape=(n_sim, n)).T
        Z = (
            self._rho
            * W
            * jnp.sqrt(1 - self._rho**2)
            * random.normal(key, shape=(n_sim, n)).T
        )

        return _sample_paths(
            self.s0, self.v0, self.mean, self.kappa, self.theta, self.sigma, dt, W, Z, n
        )


class MultiHestonModel:
    """Multivariate Heston Model

    Represent a multi-dimensional Heston model

    # Example usage:
    s0 = jnp.array([100, 100], dtype=jnp.float32)
    v0 = jnp.array([0.04, 0.04], dtype=jnp.float32)
    mean = jnp.array([0.05, 0.05], dtype=jnp.float32)
    kappa = jnp.array([2.0, 2.0], dtype=jnp.float32)
    theta = jnp.array([0.04, 0.04], dtype=jnp.float32)
    sigma = jnp.array([0.3, 0.3], dtype=jnp.float32)
    corr = jnp.array([[1.0, -0.7], [-0.7, 1.0]], dtype=jnp.float32)

    heston_model = MultiVariateHestonModel(s0, v0, mean, kappa, theta, sigma, corr, dtype=jnp.float32)
    paths = heston_model.sample_paths(seed=42, maturity=1.0, n=100, n_sim=100)
    """

    def __init__(
        self,
        s0: jax.Array,
        v0: jax.Array,
        mean: jax.Array,
        kappa: jax.Array,
        theta: jax.Array,
        sigma: jax.Array,
        corr: jax.Array,
        dtype: jnp.dtype = jnp.float32,
    ):
        """Multivariate Hestion model constructor

        Args:
            s0 (jax.Array): The initial values of the assets
            v0 (jax.Array): The initial values of the variance of the assets
            mean (jax.Array): The mean(drift term) of the assets
            kappa (jax.Array): The speed of the mean-reversion of the variance
            theta (jax.Array): The long-term mean of the variance
            sigma (jax.Array): The volatility of the variance of the assets
            corr (jax.Array): The correlation matrix of the Heston model
            dtype (Optional[jnp.dtype], optional): The dtype to use. Defaults to jnp.float32.
        """
        if not check_symmetric(corr, 1e-8):
            raise ValueError("The correlation matrix must be symmetric")

        self._dtype = dtype
        self._s0 = jnp.asarray(s0, dtype=dtype)
        self._v0 = jnp.asarray(v0, dtype=dtype)
        self._mean = jnp.asarray(mean, dtype=dtype)
        self._kappa = jnp.asarray(kappa, dtype=dtype)
        self._theta = jnp.asarray(theta, dtype=dtype)
        self._sigma = jnp.asarray(sigma, dtype=dtype)
        self._corr = jnp.asarray(corr, dtype=dtype)
        self._dim = self._s0.shape[0]

    @property
    def mean(self) -> jax.Array:
        """Returns the mean of the Heston model

        Returns:
            jax.Array: The mean of the Heston model
        """
        return self._mean

    @property
    def kappa(self) -> jax.Array:
        """Returns the speed of the mean-reversion of the variance

        Returns:
            jax.Array: The speed of the mean-reversion of the variance
        """
        return self._kappa

    @property
    def theta(self) -> jax.Array:
        """Returns the long-term mean of the variance

        Returns:
            jax.Array: The long-term mean of the variance
        """
        return self._theta

    @property
    def sigma(self) -> jax.Array:
        """Returns the volatility of the variance

        Returns:
            jax.Array: The volatility of the variance
        """
        return self._sigma

    @property
    def s0(self) -> jax.Array:
        """Returns the initial value of the asset

        Returns:
            jax.Array: The initial value of the asset
        """
        return self._s0

    @property
    def v0(self) -> jax.Array:
        """Returns the initial value of the variance of the asset

        Returns:
            jax.Array: The initial value of the variance of the asset
        """
        return self._v0

    @property
    def corr(self) -> jax.Array:
        """Returns the correlation matrix of the Heston model

        Returns:
            jax.Array: The correlation matrix of the Heston model
        """
        return self._corr

    @property
    def dtype(self) -> jax.numpy.dtype:
        """Returns the underlying dtype of the Heston model

        Returns:
            jax.numpy.dtype: The underlying dtype of the Heston model
        """
        return self._dtype

    def sample_paths(self, seed: int, maturity: float, n: int, n_sim: int) -> jax.Array:
        """Sample paths from the Multivariate Heston model

        Args:
            seed (int): The seed for the random number generator
            maturity (float): The time in years
            n (int): The number of steps
            n_sim (int): The number of simulations

        Returns:
            jax.Array: The simulated paths
        """
        key = random.PRNGKey(seed)

        dt = maturity / n

        W = random.normal(key, shape=(n_sim, n * self._dim))
        Z = random.normal(key, shape=(n_sim, n * self._dim))
        W = jnp.reshape(W, (n_sim, n, self._dim)).transpose((1, 0, 2))
        Z = jnp.reshape(Z, (n_sim, n, self._dim)).transpose((1, 0, 2))

        L = jnp.linalg.cholesky(self._corr)

        epsilon_S = W @ L
        epsilon_v = epsilon_S @ self._corr + Z @ jnp.sqrt(1 - self._corr**2)

        return _sample_paths(
            self.s0,
            self.v0,
            self.mean,
            self.kappa,
            self.theta,
            self.sigma,
            dt,
            epsilon_S,
            epsilon_v,
            n,
        )


# Helpers


def _variance_process_step(
    v: jax.Array,
    kappa: jax.Array,
    theta: jax.Array,
    sigma: jax.Array,
    dt: jax.Array,
    dZ: jax.Array,
) -> jax.Array:
    return (
        v
        + kappa * (theta - jnp.maximum(v, 0.0)) * dt
        + sigma * jnp.sqrt(jnp.maximum(v, 0.0)) * jnp.sqrt(dt) * dZ
    )


@jit
def _sample_paths(
    s0: jax.Array,
    v0: jax.Array,
    mean: jax.Array,
    kappa: jax.Array,
    theta: jax.Array,
    sigma: jax.Array,
    dt: jax.Array,
    W: jax.Array,
    Z: jax.Array,
    N: int,
) -> jax.Array:
    """Main function that simulate the path of the Heston model leveragin the
       jax.lax.while_loop function that allows to perform a loop in a jax makes
       it useful for reducing compilation times for jit-compiled functions,
       since native Python loop constructs in an @jit function are unrolled,
       leading to large XLA computations.

    Args:
        s0 (jax.Array): The initial price of the stock(s)
        v0 (jax.Array): The initial variance of the stock(s)
        mean (jax.Array): The mean of the stock(s)
        kappa (jax.Array): The speed of the mean-reversion of the variance
        theta (jax.Array): The long-term mean of the variance
        sigma (jax.Array): The volatility of the variance of the stock(s)
        dt (jax.Array): The time step
        W (jax.Array): The Weiner process of the stock(s)
        Z (jax.Array): The (correlated) Weiner process of the variance of the stock(s)
        N (int): The number of steps

    Returns:
        jax.Array: The simulated paths of the stock(s)
    """

    def init_fn(W, Z):
        W_ = W
        Z_ = Z

        S = jnp.full_like(W_, s0)
        v = jnp.full_like(W_, v0)

        return 0, S, v, W_, Z_

    def cond_fn(val):
        i, *_ = val
        return i < N

    def body_val(val):
        i, S, v, W, Z = val
        dW = W[i]
        dZ = Z[i]

        S = S.at[i + 1].set(
            S[i]
            * jnp.exp((mean - 0.5 * v[i]) * dt + jnp.sqrt(v[i]) * jnp.sqrt(dt) * dW)
        )

        v = v.at[i + 1].set(_variance_process_step(v[i], kappa, theta, sigma, dt, dZ))

        return i + 1, S, v, W, Z

    _, S, _, _, _ = jax.lax.while_loop(cond_fn, body_val, init_fn(W, Z))

    return S



import jax.numpy as jnp

import pytest

from jaxfin.price_engine.black import future_option_price, future_option_delta, future_option_gamma
from jaxfin.price_engine.utils.vect import get_vfunction

TOL = 1e-3
DTYPE = jnp.float32

def test_one_future_option_float():
    spot = jnp.array(100, dtype=DTYPE)
    expire = jnp.array(1.0, dtype=DTYPE)
    vol = jnp.array(0.3, dtype=DTYPE)
    strike = jnp.array(110, dtype=DTYPE)
    risk_free_rate = jnp.array(0.0, dtype=DTYPE)

    price = future_option_price(spot, strike, expire, vol, risk_free_rate, dtype=DTYPE)

    assert price == 8.141014

def test_one_future_option():
    spot = jnp.array([100], dtype=DTYPE)
    expire = jnp.array([1.0], dtype=DTYPE)
    vol = jnp.array([0.3], dtype=DTYPE)
    strike = jnp.array([110], dtype=DTYPE)
    risk_free_rate = jnp.array([0.0], dtype=DTYPE)

    price = future_option_price(spot, strike, expire, vol, risk_free_rate, dtype=DTYPE)

    assert price[0] == 8.141014


def test_foption_batch_calls():
    spots = jnp.array([100, 90, 80, 110, 120], dtype=DTYPE)
    expires = jnp.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=DTYPE)
    vols = jnp.array([.3, .25, .4, .2, .1], dtype=DTYPE)
    strikes = jnp.array([120, 120, 120, 120, 120], dtype=DTYPE)
    discount_rates = jnp.array([0.00, 0.00, 0.00, 0.00, 0.00], dtype=DTYPE)

    prices = future_option_price(spots, strikes, expires, vols, discount_rates,
                                  dtype=DTYPE)
    expected = jnp.array([5.440567, 1.602787, 3.140933, 5.010391, 4.7853127])

    assert jnp.isclose(prices, expected, atol=TOL).all()


def test_foption_batch_mixed():
    spots = jnp.array([100, 90, 80, 110, 120], dtype=DTYPE)
    expires = jnp.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=DTYPE)
    vols = jnp.array([.3, .25, .4, .2, .1], dtype=DTYPE)
    strikes = jnp.array([120, 75, 75, 120, 120], dtype=DTYPE)
    are_calls = jnp.array([True, False, False, True, True], dtype=jnp.bool_)
    discount_rates = jnp.array([0.00, 0.00, 0.00, 0.00, 0.00], dtype=DTYPE)

    prices = future_option_price(spots, strikes, expires, vols, discount_rates,
                                 are_calls=are_calls, dtype=DTYPE)
    expected = jnp.array([5.440567, 2.7794075, 9.942638, 5.010391, 4.7853127])

    assert jnp.isclose(prices, expected, atol=TOL).all()


def test_foption_batch_mixed_disc():
    spots = jnp.array([100, 90, 80, 110, 120], dtype=DTYPE)
    expires = jnp.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=DTYPE)
    vols = jnp.array([.3, .25, .4, .2, .1], dtype=DTYPE)
    strikes = jnp.array([120, 75, 75, 120, 120], dtype=DTYPE)
    are_calls = jnp.array([True, False, False, True, True], dtype=jnp.bool_)
    discount_rates = jnp.array([0.01, 0.0125, 0.02, 0.03, 0.05], dtype=DTYPE)

    prices = future_option_price(spots, strikes, expires, vols,
                                 are_calls=are_calls, discount_rates=discount_rates, dtype=DTYPE)
    expected = jnp.array([5.7082324, 2.5256162, 9.177393, 6.055744, 7.7550125])

    assert jnp.isclose(prices, expected, atol=TOL).all()

class TestDelta:

    def test_delta_float(self):
        spot = jnp.array(100, dtype=DTYPE)
        strike = jnp.array(120, dtype=DTYPE)
        expire = jnp.array(1, dtype=DTYPE)
        vol = jnp.array(0.3, dtype=DTYPE)
        rate = jnp.array(0.0, dtype=DTYPE)
        expected_delta = jnp.array(0.32357, dtype=DTYPE)
        expected_put_delta = jnp.array(-0.67643, dtype=DTYPE)

        call_delta = future_option_delta(spot, strike, expire, vol, rate)
        put_delta = future_option_delta(spot, strike, expire, vol, rate, are_calls=False)

        assert jnp.isclose(call_delta, expected_delta, atol=TOL)
        assert jnp.isclose(put_delta, expected_put_delta, atol=TOL)

    def test_delta(self):
        spots = jnp.array([100, 100, 100, 80, 170], dtype=DTYPE)
        strikes = jnp.array([120, 110, 120, 150, 160], dtype=DTYPE)
        expires = jnp.array([1, 1, 1, 0.5, 0.25], dtype=DTYPE)
        vols = jnp.array([0.3, 0.3, 0.2, 0.5, 0.15], dtype=DTYPE)
        rates = jnp.array([0.0, 0.0, 0.05, 0.02, 0.01], dtype=DTYPE)
        dividends = jnp.array([0.0, 0.0, 0.0, 0.0, 0.0], dtype=DTYPE)
        call_flags = jnp.array([True, True, True, True, True], dtype=jnp.bool_)
        put_flags = jnp.array([False, False, False, False, False], dtype=jnp.bool_)
        e_call_deltas = jnp.array([0.32357, 0.43340, 0.28024, 0.05777, 0.81046], dtype=DTYPE)
        e_put_deltas = jnp.array([-0.67643, -0.56659, -0.71975, -0.94222, -0.18953], dtype=DTYPE)
        vmap_delta = get_vfunction(future_option_delta, spots, strikes, expires, vols, rates, dividends, are_calls=put_flags)

        call_deltas = vmap_delta(spots, strikes, expires, vols, rates, dividends, call_flags)
        put_deltas = vmap_delta(spots, strikes, expires, vols, rates, dividends, put_flags)

        assert jnp.allclose(call_deltas, e_call_deltas, atol=TOL)
        assert jnp.allclose(put_deltas, e_put_deltas, atol=TOL)


@pytest.mark.parametrize("spot, strike, expire, vol, rate, expected_gamma",
                            [(100, 120, 1, 0.3, 0.0, 0.01197),
                            (100, 110, 1, 0.3, 0.0, 0.013112),
                            (100, 120, 1, 0.2, 0.01, 0.01523321),
                            (80, 150, 0.5, 0.5, 0.02, 0.00417),
                            (170, 160, 0.25, 0.15, 0.01, 0.02136)
                            ])
class TestGamma:

    def test_gamma_bs(self, spot, strike, expire, vol, rate, expected_gamma):
        spot = jnp.array(spot, dtype=DTYPE)
        strike = jnp.array(strike, dtype=DTYPE)
        expire = jnp.array(expire, dtype=DTYPE)
        vol = jnp.array(vol, dtype=DTYPE)
        rate = jnp.array(rate, dtype=DTYPE)
        dividends = jnp.array(0.0, dtype=DTYPE)
        put_flag = jnp.array(False, dtype=jnp.bool_)
        expected_gamma = jnp.array(expected_gamma, dtype=DTYPE)
        call_gamma = future_option_gamma(spot, strike, expire, vol, rate, dividends)
        put_gamma = future_option_gamma(spot, strike, expire, vol, rate, dividends, are_calls=put_flag)

        assert jnp.isclose(call_gamma, expected_gamma, atol=TOL).all()
        assert jnp.isclose(put_gamma, call_gamma, atol=TOL).all()



import jax.numpy as jnp

from jaxfin.price_engine.black_scholes import european_price, delta_european, gamma_european, theta_european, rho_european, vega_european
from jaxfin.price_engine.utils.vect import get_vfunction

import logging

import pytest

TOL = 1e-3
DTYPE = jnp.float32

logging.basicConfig(level=logging.INFO)

def test_one_vanilla():
    dtype = jnp.float32
    spot = jnp.array([100], dtype=dtype)
    expire = jnp.array([1.0], dtype=dtype)
    vol = jnp.array([0.3], dtype=dtype)
    strike = jnp.array([110], dtype=dtype)
    risk_free_rate = jnp.array([0.0], dtype=dtype)

    price = european_price(spot, strike, expire, vol, risk_free_rate, dtype=dtype)

    assert price[0] == 8.141014


def test_vanilla_batch_calls():
    spots = jnp.array([100, 90, 80, 110, 120], dtype=DTYPE)
    expires = jnp.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=DTYPE)
    vols = jnp.array([.3, .25, .4, .2, .1], dtype=DTYPE)
    strikes = jnp.array([120, 120, 120, 120, 120], dtype=DTYPE)
    discount_rates = jnp.array([0.00, 0.00, 0.00, 0.00, 0.00], dtype=DTYPE)

    prices = european_price(spots, strikes, expires, vols, discount_rates, dtype=DTYPE)
    expected = jnp.array([5.440567, 1.602787, 3.140933, 5.010391, 4.7853127])

    assert jnp.isclose(prices, expected, atol=TOL).all()


def test_vanilla_batch_mixed():
    spots = jnp.array([100, 90, 80, 110, 120], dtype=DTYPE)
    expires = jnp.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=DTYPE)
    vols = jnp.array([.3, .25, .4, .2, .1], dtype=DTYPE)
    strikes = jnp.array([120, 75, 75, 120, 120], dtype=DTYPE)
    are_calls = jnp.array([True, False, False, True, True], dtype=jnp.bool_)
    discount_rates = jnp.array([0.00, 0.00, 0.00, 0.00, 0.00], dtype=DTYPE)

    prices = european_price(spots, strikes, expires, vols, discount_rates,
                            are_calls=are_calls, dtype=DTYPE)
    expected = jnp.array([5.440567, 2.7794075, 9.942638, 5.010391, 4.7853127])

    assert jnp.isclose(prices, expected, atol=TOL).all()


def test_vanilla_batch_mixed_disc():
    spots = jnp.array([100, 90, 80, 110, 120], dtype=DTYPE)
    expires = jnp.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=DTYPE)
    vols = jnp.array([.3, .25, .4, .2, .1], dtype=DTYPE)
    strikes = jnp.array([120, 75, 75, 120, 120], dtype=DTYPE)
    are_calls = jnp.array([True, False, False, True, True], dtype=jnp.bool_)
    discount_rates = jnp.array([0.01, 0.0125, 0.02, 0.03, 0.05], dtype=DTYPE)

    prices = european_price(spots, strikes, expires, vols,
                            are_calls=are_calls, discount_rates=discount_rates, dtype=DTYPE)
    expected = jnp.array([5.714386,  2.5328674, 9.19194, 6.155064, 8.165947])

    assert jnp.isclose(prices, expected, atol=TOL).all()

### Test Greeks Calculations ###

@pytest.mark.parametrize("spot, strike, expire, vol, rate, e_call_delta, e_put_delta",
                         [(100, 120, 1, 0.3, 0.0, 0.32357, -0.67643),
                          (100, 110, 1, 0.3, 0.0, 0.43341, -0.56659),
                          (100, 120, 1, 0.2, 0.05, 0.28719, -0.71281),
                          (80, 150, 0.5, 0.5, 0.02, 0.05787, -0.94213),
                          (170, 160, 0.25, 0.15, 0.01, 0.81034, -0.18966)
                        ])
class TestDelta:

    def test_delta_bs(self, spot, strike, expire, vol, rate, e_call_delta, e_put_delta):
        spot = jnp.array(spot, dtype=DTYPE)
        strike = jnp.array(strike, dtype=DTYPE)
        expire = jnp.array(expire, dtype=DTYPE)
        vol = jnp.array(vol, dtype=DTYPE)
        rate = jnp.array(rate, dtype=DTYPE)
        put_flag = jnp.array(False, dtype=jnp.bool_)
        e_call_delta = jnp.array(e_call_delta, dtype=DTYPE)
        call_delta = delta_european(spot, strike, expire, vol, rate)
        put_delta = delta_european(spot, strike, expire, vol, rate, are_calls=put_flag)

        assert jnp.isclose(call_delta, e_call_delta, atol=TOL).all()
        assert jnp.isclose(put_delta, e_put_delta, atol=TOL).all()
        assert jnp.isclose(call_delta - 1.0, put_delta, atol=TOL).all()



class TestDeltaBatch:

    def test_delta_bs_scalar(self):
        spot = jnp.array(100, dtype=DTYPE)
        strike = jnp.array(120, dtype=DTYPE)
        expire = jnp.array(1, dtype=DTYPE)
        vol = jnp.array(0.3, dtype=DTYPE)
        rate = jnp.array(0.0, dtype=DTYPE)
        expected_delta = jnp.array(0.32357, dtype=DTYPE)
        expected_put_delta = jnp.array(-0.67643, dtype=DTYPE)

        logging.info(f"Testing with spot={spot}, strike={strike}, expire={expire}, vol={vol}, rate={rate}")

        call_delta = delta_european(spot, strike, expire, vol, rate)
        put_delta = delta_european(spot, strike, expire, vol, rate, are_calls=False)

        assert jnp.isclose(call_delta, expected_delta, atol=TOL)
        assert jnp.isclose(put_delta, expected_put_delta, atol=TOL)
        assert jnp.isclose(call_delta - 1.0, put_delta, atol=TOL)

    def test_delta_bs_batch(self):
        spots = jnp.array([100, 90, 80, 110, 120], dtype=DTYPE)
        expires = jnp.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=DTYPE)
        vols = jnp.array([.3, .25, .4, .2, .1], dtype=DTYPE)
        strikes = jnp.array([120, 120, 120, 120, 120], dtype=DTYPE)
        discount_rates = jnp.array([0.00, 0.00, 0.00, 0.00, 0.00], dtype=DTYPE)
        call_flag = jnp.array([True, True, True, True, True], dtype=jnp.bool_)

        vmap_delta = get_vfunction(delta_european, spots, strikes, expires, vols, discount_rates, call_flag)

        logging.info(f"Testing with spots={spots}, strikes={strikes}, expires={expires}, vols={vols}, rates={discount_rates}")

        call_delta = vmap_delta(spots, strikes, expires, vols, discount_rates, call_flag)
        put_flag = jnp.array([False, False, False, False, False], dtype=jnp.bool_)
        put_delta = vmap_delta(spots, strikes, expires, vols, discount_rates, put_flag)

        expected_call = jnp.array([0.323570, 0.152509, 0.207919, 0.36879, 0.51993])
        expected_put = jnp.array([-0.676429, -0.847490, -0.79208, -0.63120, -0.480061])

        assert jnp.isclose(call_delta, expected_call, atol=TOL).all()
        assert jnp.isclose(put_delta, expected_put, atol=TOL).all()
        assert jnp.isclose(call_delta - 1.0, put_delta, atol=TOL).all()

@pytest.mark.parametrize("spot, strike, expire, vol, rate, expected_gamma",
                            [(100, 120, 1, 0.3, 0.0, 0.01198),
                            (100, 110, 1, 0.3, 0.0, 0.01311),
                            (100, 120, 1, 0.2, 0.05, 0.01704),
                            (80, 150, 0.5, 0.5, 0.02, 0.00409),
                            (170, 160, 0.25, 0.15, 0.01, 0.02126)
                            ])
class TestGamma:

    def test_gamma_bs(self, spot, strike, expire, vol, rate, expected_gamma):
        spot = jnp.array(spot, dtype=DTYPE)
        strike = jnp.array(strike, dtype=DTYPE)
        expire = jnp.array(expire, dtype=DTYPE)
        vol = jnp.array(vol, dtype=DTYPE)
        rate = jnp.array(rate, dtype=DTYPE)
        gamma = gamma_european(spot, strike, expire, vol, rate)

        assert jnp.isclose(gamma, expected_gamma, atol=TOL).all()

class TestGammaBatch:

    def test_gamma_bs_vectorized(self):
        spots = jnp.array([100, 100, 100, 80, 170], dtype=DTYPE)
        strikes = jnp.array([120, 110, 120, 150, 160], dtype=DTYPE)
        expires = jnp.array([1, 1, 1, 0.5, 0.25], dtype=DTYPE)
        vols = jnp.array([0.3, 0.3, 0.2, 0.5, 0.15], dtype=DTYPE)
        rates = jnp.array([0.0, 0.0, 0.05, 0.02, 0.01], dtype=DTYPE)
        expected_gammas = jnp.array([0.01198, 0.01311, 0.01704, 0.00409, 0.02126], dtype=DTYPE)
        vmap_gamma = get_vfunction(gamma_european, spots, strikes, expires, vols, rates)

        gammas = vmap_gamma(spots, strikes, expires, vols, rates)

        assert jnp.allclose(gammas, expected_gammas, atol=TOL)

@pytest.mark.parametrize("spot, strike, expire, vol, rate, e_call_theta, e_put_theta",
                            [(100, 120, 1, 0.3, 0.0, 5.38894, 5.38894),
                            (100, 110, 1, 0.3, 0.0, 5.90058, 5.90058),
                            (100, 120, 1, 0.2, 0.05, 4.68097, -1.02641),
                            (80, 150, 0.5, 0.5, 0.02, 3.35533, 0.38519),
                            (170, 160, 0.25, 0.15, 0.01, 8.17193, 6.57593)
                            ])
class TestTheta:

    def test_theta_bs(self, spot, strike, expire, vol, rate, e_call_theta, e_put_theta):
        spot = jnp.array(spot, dtype=DTYPE)
        strike = jnp.array(strike, dtype=DTYPE)
        expire = jnp.array(expire, dtype=DTYPE)
        vol = jnp.array(vol, dtype=DTYPE)
        rate = jnp.array(rate, dtype=DTYPE)
        put_flag = jnp.array(False, dtype=jnp.bool_)
        e_call_theta = jnp.array(e_call_theta, dtype=DTYPE)
        e_put_theta = jnp.array(e_put_theta, dtype=DTYPE)
        theta = theta_european(spot, strike, expire, vol, rate)
        p_theta = theta_european(spot, strike, expire, vol, rate, are_calls=put_flag)

        assert jnp.isclose(theta, e_call_theta, atol=TOL).all()
        assert jnp.isclose(p_theta, e_put_theta, atol=TOL).all()

@pytest.mark.parametrize("spot, strike, expire, vol, rate, e_call_rho, e_put_rho",
                            [(100, 120, 1, 0.3, 0.0, 26.91644, -93.08356),
                            (100, 110, 1, 0.3, 0.0, 35.19993, -74.80007),
                            (100, 120, 1, 0.2, 0.05, 25.47168, -88.67585),
                            (80, 150, 0.5, 0.5, 0.02, 2.00656, -72.24718),
                            (170, 160, 0.25, 0.15, 0.01, 31.49509, -8.40503)
                            ]) 
class TestRho:

    def test_rho_bs(self, spot, strike, expire, vol, rate, e_call_rho, e_put_rho):
        spot = jnp.array(spot, dtype=DTYPE)
        strike = jnp.array(strike, dtype=DTYPE)
        expire = jnp.array(expire, dtype=DTYPE)
        vol = jnp.array(vol, dtype=DTYPE)
        rate = jnp.array(rate, dtype=DTYPE)
        put_flag = jnp.array(False, dtype=jnp.bool_)
        e_call_rho = jnp.array(e_call_rho, dtype=DTYPE)
        call_rho = rho_european(spot, strike, expire, vol, rate)
        put_rho = rho_european(spot, strike, expire, vol, rate, are_calls=put_flag)

        assert jnp.isclose(call_rho, e_call_rho, atol=TOL).all()
        assert jnp.isclose(put_rho, e_put_rho, atol=TOL).all()

@pytest.mark.parametrize("spot, strike, expire, vol, rate, expected_vega",
                            [(100, 120, 1, 0.3, 0.0, 35.92629),
                            (100, 110, 1, 0.3, 0.0, 39.33717),
                            (100, 120, 1, 0.2, 0.05, 34.07384),
                            (80, 150, 0.5, 0.5, 0.02, 6.55014),
                            (170, 160, 0.25, 0.15, 0.01, 23.04042)
                            ])
class TestVega:
    
        def test_vega_bs(self, spot, strike, expire, vol, rate, expected_vega):
            spot = jnp.array(spot, dtype=DTYPE)
            strike = jnp.array(strike, dtype=DTYPE)
            expire = jnp.array(expire, dtype=DTYPE)
            vol = jnp.array(vol, dtype=DTYPE)
            rate = jnp.array(rate, dtype=DTYPE)
            vega = vega_european(spot, strike, expire, vol, rate)
    
            assert jnp.isclose(vega, expected_vega, atol=TOL).all()



import jax.numpy as jnp
from jaxfin.models.heston.heston import UnivHestonModel, MultiHestonModel

SEED: int = 42

class TestUnivHestonModel():

    def test_init(self):
        s0 = 100
        v0 = 0.2
        mean = 0.2
        kappa = 2.0
        theta = 0.25
        sigma = 0.3
        rho = -0.7
        dtype = jnp.float32

        heston_model = UnivHestonModel(s0, v0, mean, kappa, theta, sigma, rho, dtype=dtype)

        assert heston_model.mean == mean
        assert heston_model.kappa == kappa
        assert heston_model.theta == theta
        assert heston_model.sigma == sigma
        assert heston_model.rho == rho
        assert heston_model.dtype == dtype
        assert heston_model.s0 == s0
        assert heston_model.v0 == v0

    def test_sample_paths(self):
        s0 = jnp.array(100, dtype=jnp.float32)
        v0 = jnp.array(0.2, dtype=jnp.float32)
        mean = jnp.array(0.2, dtype=jnp.float32)
        kappa = jnp.array(2.0, dtype=jnp.float32)
        theta = jnp.array(0.25, dtype=jnp.float32)
        sigma = jnp.array(0.3, dtype=jnp.float32)
        rho = jnp.array(-0.7, dtype=jnp.float32)

        heston_model = UnivHestonModel(s0, v0, mean, kappa, theta, sigma, rho, dtype=jnp.float32)
        paths = heston_model.sample_paths(seed=SEED, maturity=1.0, n=100, n_sim=100)

        assert paths.shape == (100, 100)


class TestMultiHestonModel():

    def test_init(self):
        s0 = jnp.array([100, 100], dtype=jnp.float32)
        v0 = jnp.array([0.2, 0.2], dtype=jnp.float32)
        mean = jnp.array([0.2, 0.2], dtype=jnp.float32)
        kappa = jnp.array([2.0, 2.0], dtype=jnp.float32)
        theta = jnp.array([0.25, 0.25], dtype=jnp.float32)
        sigma = jnp.array([0.3, 0.3], dtype=jnp.float32)
        corr = jnp.array([[1.0, 0.5], [0.5, 1.0]], dtype=jnp.float32)
        dtype = jnp.float32

        heston_model = MultiHestonModel(s0, v0, mean, kappa, theta, sigma, corr, dtype=dtype)

        assert jnp.all(heston_model.mean == mean)
        assert jnp.all(heston_model.kappa == kappa)
        assert jnp.all(heston_model.theta == theta)
        assert jnp.all(heston_model.sigma == sigma)
        assert jnp.all(heston_model.corr == corr)
        assert heston_model.dtype == dtype
        assert jnp.all(heston_model.s0 == s0)
        assert jnp.all(heston_model.v0 == v0)

    def test_sample_paths(self):
        s0 = jnp.array([100, 100], dtype=jnp.float32)
        v0 = jnp.array([0.2, 0.2], dtype=jnp.float32)
        mean = jnp.array([0.2, 0.2], dtype=jnp.float32)
        kappa = jnp.array([2.0, 2.0], dtype=jnp.float32)
        theta = jnp.array([0.25, 0.25], dtype=jnp.float32)
        sigma = jnp.array([0.3, 0.3], dtype=jnp.float32)
        corr = jnp.array([[1.0, 0.5], [0.5, 1.0]], dtype=jnp.float32)

        heston_model = MultiHestonModel(s0, v0, mean, kappa, theta, sigma, corr, dtype=jnp.float32)
        paths = heston_model.sample_paths(seed=SEED, maturity=1.0, n=100, n_sim=100)

        assert paths.shape == (100, 100, 2)



"""
Common standard normal related functions used in multiple submodules in the price_engine module
"""

import jax
import jax.numpy as jnp
from jax.scipy.special import erf
from jax.scipy.stats.norm import pdf

_SQRT_2 = jnp.sqrt(2.0)


@jax.jit
def cum_normal(x):
    """
    Cumulative normal distribution function

    :param x: Input value
    :return: Cumulative normal distribution value
    """
    return (erf(x / _SQRT_2) + 1) / 2


@jax.jit
def density_normal(x):
    """
    Normal distribution function

    :param x: Input value
    :return: Normal distribution value
    """
    return pdf(x)



import jax.numpy as jnp

from jaxfin.models.gbm import UnivGeometricBrownianMotion, MultiGeometricBrownianMotion

SEED: int = 42

class TestUnivGBM:

    def test_init(self):
        s0 = 10
        mean = 0.1
        sigma = 0.3
        dtype = jnp.float32
        gbm = UnivGeometricBrownianMotion(s0, mean, sigma, dtype)

        assert gbm.mean == mean
        assert gbm.sigma == sigma
        assert gbm.dtype == dtype
        assert gbm.s0 == s0

    def test_sim_paths_shape(self):
        s0 = 10
        mean = 0.1
        sigma = 0.3
        dtype = jnp.float32
        gbm = UnivGeometricBrownianMotion(s0, mean, sigma, dtype)

        stock_paths = gbm.sample_paths(SEED, 1.0, 52, 100)

        assert stock_paths.shape == (52, 100)


class TestMultiGBM:

    def test_init(self):
        s0 = jnp.array([10, 12])
        mean = jnp.array([0.1, 0.0])
        sigma = jnp.array([0.3, 0.5])
        corr = jnp.array([[1, 0.1], [0.1, 1]])
        dtype = jnp.float32
        gbm = MultiGeometricBrownianMotion(s0, mean, sigma, corr, dtype)

        assert jnp.array_equal(gbm.mean, mean)
        assert jnp.array_equal(gbm.sigma, sigma)
        assert jnp.array_equal(gbm.corr, corr)
        assert gbm.dtype == dtype
        assert jnp.array_equal(gbm.s0, s0)
        assert gbm.dimension == 2


    def test_sample_path(self):
        s0 = jnp.array([10, 12])
        mean = jnp.array([0.1, 0.0])
        sigma = jnp.array([0.3, 0.5])
        corr = jnp.array([[1, 0.1], [0.1, 1]])
        dtype = jnp.float32
        gbm = MultiGeometricBrownianMotion(s0, mean, sigma, corr, dtype)
        sample_path = gbm.sample_paths(SEED, 1.0, 52, 100)

        assert sample_path.shape == (52, 100, 2)
